import json
import tarfile
from pathlib import Path
import tempfile
import datetime

from .eos import mv_from_eos, mv_to_eos
from .sim_data import build_input_file
from .default_tracker import get_default_tracker

temp    = tempfile.TemporaryDirectory()
tempdir = Path(temp.name).resolve()
eosdir  = '/eos/user/d/ddicroce/xboinc/'

def timestamp(ms=False):
    ms = -3 if ms else -7
    return datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S.%f")[:ms]

class SubmitJobs:

    def __init__(self, user, study):
        if '__' in study:
            raise ValueError("The character sequence '__' is not allowed in 'study'!")
        self._username = user
        self._studyname = study
        self._submitfile = f"{self._studyname}__{timestamp()}.tar.gz"
        self._json_files = []
        self._bin_files = []
        tempdir.mkdir(parents=True, exist_ok=True)

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        self._submit()
        temp.cleanup()

    def add(self, *, job_name, num_turns, line, particles, checkpoint_every=-1):
        if '__' in job_name:
            raise ValueError("The character sequence '__' is not allowed in 'job_name'!")
        filename = f"{self._username}__{timestamp(ms=True)}"
        json_file = Path(tempdir, f"{filename}.json")
        bin_file  = Path(tempdir, f"{filename}.bin")
        json_dict = {'study': self._studyname, 'user': self._username, 'job_name': job_name}
        with open(json_file, 'w') as fid:
            json.dump(json_dict, fid)
        build_input_file(name=bin_file, num_turns=num_turns, line=line,
                         particles=particles, checkpoint_every=checkpoint_every)
        self._json_files += [json_file]
        self._bin_files  += [bin_file]

    def _submit(self):
        with tarfile.open(tempdir / self._submitfile, "w:gz") as tar:
            for thisfile in self._json_files + self._bin_files:
                tar.add(thisfile, arcname=thisfile.name)
        mv_to_eos(tempdir / self._submitfile, eosdir)
        # TODO: check that tar contains all files
        for thisfile in self._json_files + self._bin_files:
            thisfile.unlink()




from .generate_executable_source import generate_executable_source, generate_executable
from .general import _pkg_root
from .sim_data import build_input_file, read_output_file
#from .submission import sub_config, prepare_job, prepare_study, submit_study
from .submit import SubmitJobs

__version__ = '0.0.3'




import xtrack as xt
import xfields as xf
import xobjects as xo
import xcoll as xc



def get_default_tracker():
    """
    Returns a default tracker object.
    """

    # Dummy line containing all supported element types
    default_line = xt.Line(elements=[
        xt.Drift(length=1.0),
        xt.Multipole(knl=[0]),
        xt.SimpleThinBend(),
        xt.SimpleThinQuadrupole(),
        xt.ReferenceEnergyIncrease(),
        xt.Cavity(),
        xt.XYShift(),
        xt.Elens(),
        xt.Wire(),
        xt.SRotation(),
        xt.XRotation(),
        xt.YRotation(),
        xt.ZetaShift(),
        xt.RFMultipole(knl=[0], pn=[0]),
        xt.DipoleEdge(),
        xt.Exciter(nsamples=1),
        xt.LineSegmentMap(),
        # xt.EnergyChange(), # not working!!!
        xf.BeamBeamBiGaussian2D(
            other_beam_Sigma_11=1.,
            other_beam_Sigma_33=1.,
            other_beam_num_particles=0.,
            other_beam_q0=1.,
            other_beam_beta0=1.,
        ),
        xf.BeamBeamBiGaussian3D(
            slices_other_beam_zeta_center=[0],
            slices_other_beam_num_particles=[0],
            phi=0.,
            alpha=0,
            other_beam_q0=1.,
            slices_other_beam_Sigma_11=[1],
            slices_other_beam_Sigma_12=[0],
            slices_other_beam_Sigma_22=[0],
            slices_other_beam_Sigma_33=[1],
            slices_other_beam_Sigma_34=[0],
            slices_other_beam_Sigma_44=[0],
        ),
        xc.BlackAbsorber(length=1),
        xc.EverestCollimator(length=1, material=xc.materials.Silicon),
        xc.EverestCrystal(length=1, material=xc.materials.SiliconCrystal)
        #xf.SpaceChargeBiGaussian( # Not working!
        #    longitudinal_profile=xf.LongitudinalProfileQGaussian(
        #        number_of_particles=0, sigma_z=1)
        #),
    ])

    _context = xo.ContextCpu()
    default_line.build_tracker(_context=_context, compile=True,
                               use_prebuilt_kernels=False)
    default_config_hash = default_line.tracker._hashable_config()
    return default_line.tracker, default_config_hash



import xobjects as xo
from .default_tracker import get_default_tracker
from .sim_data import SimState, SimConfig
from .general import _pkg_root

from pathlib import Path
import shutil
import subprocess

insert_in_all_files = """
#include <stdio.h>
#ifndef NULL
    #define NULL 0
#endif
"""

def generate_executable_source(write_source_files=True,
                                       _context=None):

    assert _context is None
    assert write_source_files

    sim_config_sources = [
        insert_in_all_files,
        xo.specialize_source(SimState._XoStruct._gen_c_api().source,
                                    specialize_for='cpu_serial'),
        xo.specialize_source(SimConfig._gen_c_api().source,
                                    specialize_for='cpu_serial'),
    ]

    sim_config_h = '\n'.join(sim_config_sources)

    default_tracker, default_config_hash = get_default_tracker()
    track_kernel = default_tracker.track_kernel[default_config_hash]
    xtrack_tracker_h = (
        insert_in_all_files + track_kernel.specialized_source)

    with open(_pkg_root.joinpath('executable_src/main.c'), 'r') as fid:
        main_c = insert_in_all_files + fid.read()

    dct_sources = {
        'sim_config.h': sim_config_h,
        'xtrack_tracker.h': xtrack_tracker_h,
        'main.c': main_c,
    }

    for nn, vv in dct_sources.items():
        with open(nn, 'w') as fid:
            fid.write(vv)

    return dct_sources


def generate_executable(keep_source=False):
    
    main    = Path.cwd() / "main.c"
    config  = Path.cwd() / "sim_config.h"
    tracker = Path.cwd() / "xtrack_tracker.h"
    if not main.exists() or not config.exists() or not tracker.exists():
        source_files = generate_executable_source()
    
    if shutil.which("gcc") is not None:
        compiler = "gcc"
    elif shutil.which("clang") is not None:
        compiler = "clang"
    else:
        raise RuntimeError("Neither clang or gcc are found. Install a C compiler.")
    
    cmd = subprocess.run([compiler, 'main.c', '-o', 'xboinc_executable', '-lm'],
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if cmd.returncode != 0:
        stderr = cmd.stderr.decode('UTF-8').strip().split('\n')
        raise RuntimeError(f"Compilation failed. Stderr:\n {stderr}")

    if not keep_source:
        main.unlink()
        config.unlink()
        tracker.unlink()



import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt

from .default_tracker import get_default_tracker

_default_tracker, _default_config_hash = get_default_tracker()


class SimState(xo.HybridClass):
    _xofields = {
        'particles': xp.Particles._XoStruct,
        'i_turn': xo.Int64,
        'size': xo.Int64,
    }


class SimConfig(xo.Struct):
    tracker_data = _default_tracker._tracker_data_cache[_default_config_hash]
    line_metadata = xo.Ref(tracker_data._element_ref_data.__class__)
    num_turns = xo.Int64
    num_elements = xo.Int64
    checkpoint_every = xo.Int64
    sim_state = xo.Ref(SimState._XoStruct)


def build_input_file(name, num_turns, line, particles, checkpoint_every=-1):
    # Assemble data structure
    simbuf = xo.ContextCpu().new_buffer()
    sim_config = SimConfig(_buffer=simbuf)
    default_tracker = _default_tracker
    line.build_tracker(_context=simbuf.context, _buffer=simbuf,
                        track_kernel=default_tracker.track_kernel)
    sim_state = SimState(_buffer=simbuf, particles=particles, i_turn=0)
    if _default_config_hash not in line.tracker._tracker_data_cache:
        raise RuntimeError('Tracker data for default config not found')
    tracker_data = line.tracker._tracker_data_cache[_default_config_hash]
    sim_config.line_metadata = tracker_data._element_ref_data
    sim_config.num_turns = num_turns
    sim_config.num_elements = len(line.element_names)
    sim_config.checkpoint_every = checkpoint_every
    sim_config.sim_state = sim_state._xobject
    sim_state.size = sim_state._xobject._size # store size of sim_state
    assert sim_config._offset == 0
    assert sim_config._fields[0].offset == 0
    # Write sim buffer to file
    with open(name, 'wb') as fid:
        fid.write(simbuf.buffer.tobytes())
    return sim_config


def read_output_file(filename):
    with open(filename, 'rb') as fid:
        state_bytes = fid.read()
    buffer_out = xo.ContextCpu().new_buffer(capacity=len(state_bytes))
    buffer_out.buffer[:] = np.frombuffer(state_bytes, dtype=np.int8)
    sim_state_xobject = SimState._XoStruct._from_buffer(buffer=buffer_out, offset=0)
    sim_state_out = SimState(_xobject=sim_state_xobject)
    return sim_state_out



import os, subprocess
from pathlib import Path
from time import sleep
import random

EOS_MGM_URL = 'root://eosuser.cern.ch'
eos_env = {**os.environ, 'EOS_MGM_URL': EOS_MGM_URL}

def _eos_path(path):
    path = Path(path).resolve()
    return path.as_posix().replace('home-','user/')

def eos_exists(path):
    try:
        path = Path(path).resolve()
        cmd = subprocess.run(['eos', 'find', '-name', path.name, _eos_path(path.parent)],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=eos_env)
        if cmd.returncode != 0 or cmd.stdout == b'':
            return False
        return Path(cmd.stdout.decode('UTF-8').strip()).resolve().stat().st_size!=0
    except Exception as e:
        print(f"Failed eos_exists for {path}!", e)
        return False

def eos_rm(path):
    try:
        path = Path(path).resolve()
        cmd = subprocess.run(['eos', 'rm', _eos_path(path)], stdout=subprocess.PIPE, env=eos_env)
        return cmd
    except Exception as e:
        print(f"Failed eos_rm for {path}!", e)

def mv_from_eos(source, target, maximum_trials=10, wait=2.7):
    try:
        target = Path(target).resolve()
        for i in range(maximum_trials):
            cmd = subprocess.run(['xrdcp', '--cksum', 'adler32', _eos_path(source), target.as_posix()])
            if cmd.returncode == 0 and target.exists() and target.stat().st_size!=0:
                eos_rm(source)
                return cmd
            if i != maximum_trials-1:
                print(f"Failed to move {str(source)} to {str(target)}!\nRetrying ({i})..")
                sleep(abs(wait+random.normalvariate(0, 0.1*wait)))
        print(f"Failed to move {str(source)} to {str(target)}!\nGiving up.")
        return cmd  
    except Exception as e:
        print(f"Failed to move {str(source)} to {str(target)}!", e)

def mv_to_eos(source, target, maximum_trials=10, wait=2.7):
    try:
        source = Path(source).resolve()
        target = Path(target, source.name).resolve()
        for i in range(maximum_trials):
            cmd = subprocess.run(['xrdcp', '--cksum', 'adler32', source.as_posix(), _eos_path(target)])
            if cmd.returncode == 0 and eos_exists(target):
                source.unlink()
                return cmd
            if i != maximum_trials-1:
                print(f"Failed to move {str(source)} to {str(target)}!\nRetrying ({i})..")
                sleep(abs(wait+random.normalvariate(0, 0.1*wait)))
        print(f"Failed to move {str(source)} to {str(target)}!\nGiving up.")
        return cmd
    except Exception as e:
        print(f"Failed to move {str(source)} to {str(target)}!", e)

